@@ -27,6 +27,8 @@ PACKAGE_NAME = "Polymer"
 
 
 """ Polymer.py - Manage parallel tasks
+     Copyright (C) 2023 David Michael Pennington at Starbucks Coffee Company
+     Copyright (C) 2021-2023 David Michael Pennington
      Copyright (C) 2021-2022 David Michael Pennington
      Copyright (C) 2020-2021 David Michael Pennington at Cisco Systems
      Copyright (C) 2019      David Michael Pennington at ThousandEyes
@@ -53,18 +55,21 @@ class SharedCounter(object):
     http://eli.thegreenplace.net/2012/01/04/shared-counter-with-pythons-multiprocessing/
     """
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def __init__(self, n=0):
         self.count = multiprocessing.Value("i", n)
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def increment(self, n=1):
         """ Increment the counter by n (default = 1) """
         with self.count.get_lock():
             self.count.value += n
 
     @property
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def value(self):
         """ Return the value of the counter """
         return self.count.value
@@ -94,7 +99,8 @@ class py3_mp_queue(mpq.Queue):
     # NOTE This is the new implementation based on:
     #    https://github.com/vterron/lemon/blob/master/util/queue.py
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def __init__(self, *args, **kwargs):
         # Use ctx argument if using Python3.4+
         super(py3_mp_queue, self).__init__(
@@ -104,7 +110,8 @@ class py3_mp_queue(mpq.Queue):
         if not sys.platform == "darwin":
             self.size = SharedCounter(0)
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def put(self, *args, **kwargs):
         # Infinite recursion possible if we don't use super() here
         try:
@@ -116,7 +123,8 @@ class py3_mp_queue(mpq.Queue):
         if not sys.platform == "darwin":
             self.size.increment(1)
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def get(self, *args, **kwargs):
         if not sys.platform == "darwin":
             self.size.increment(-1)
@@ -127,7 +135,8 @@ class py3_mp_queue(mpq.Queue):
         except Empty:
             return {}
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def qsize(self):
         """ Reliable implementation of multiprocessing.Queue.qsize() """
         if sys.platform != "darwin":
@@ -135,7 +144,8 @@ class py3_mp_queue(mpq.Queue):
         else:
             raise NotImplementedError
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def empty(self):
         """ Reliable implementation of multiprocessing.Queue.empty() """
         if sys.platform != "darwin":
@@ -143,7 +153,8 @@ class py3_mp_queue(mpq.Queue):
         else:
             raise NotImplementedError
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def clear(self):
         """ Remove all elements from the Queue. """
         while not self.empty():
@@ -153,7 +164,8 @@ class py3_mp_queue(mpq.Queue):
 class Worker(object):
     """multiprocessing worker"""
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def __init__(self, w_id, todo_q, done_q, default_sleep=0.00001):
         assert isinstance(default_sleep, int) or isinstance(default_sleep, float)
         color_init()
@@ -183,7 +195,8 @@ class Worker(object):
             # time.sleep(self.cycle_sleep)
             self.task = None
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def done_q_send(self, msg_dict):
         """Send message dicts through done_q, and throw explicit errors for
         pickle problems"""
@@ -282,7 +295,8 @@ class Worker(object):
                     )
                 )
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def invalid_dict_pickle_keys(self, msg_dict) -> list:
         """
         Return a list of keys that can't be pickled.  Return [] if
@@ -306,7 +320,8 @@ class Worker(object):
 
         return no_pickle_keys
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def invalid_obj_pickle_attrs(self, thisobj):
         no_pickle_attrs = []
         for attr, val in vars(thisobj).items():
@@ -320,7 +335,8 @@ class Worker(object):
                 no_pickle_attrs.append(attr)  # This attr is unpicklable
         return no_pickle_attrs
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def message_loop(self, todo_q, done_q):
         """Loop through messages and execute tasks"""
         t_msg = {}
@@ -339,6 +355,7 @@ class Worker(object):
                     # Update the sleep time with latest recommendations
                     self.cycle_sleep = self.task.worker_loop_delay
 
+
                     # Assign the result of task.run() to task.result
                     self.task.result = self.task.run()
                     self.task.task_stop = time.time()  # Seconds since epoch
@@ -372,7 +389,8 @@ class Worker(object):
 
 class TaskMgrStats(object):
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def __init__(self, worker_count, log_interval=5, hot_loop=False):
 
         assert isinstance(worker_count, int)
@@ -387,19 +405,22 @@ class TaskMgrStats(object):
         self.exec_times = []
         self.queue_times = []
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def reset_stats(self):
         self.stats_start = time.time()
         self.exec_times = []
         self.queue_times = []
 
     @property
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def time_delta(self):
         return time.time() - self.stats_start
 
     @property
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def is_log_time(self):
         """
         Return True if it's time to log
@@ -409,7 +430,8 @@ class TaskMgrStats(object):
         return False
 
     @property
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def log_stats_message(self):
         """
         Build log message strings and reset the stats
@@ -457,7 +479,8 @@ class TaskMgr(object):
     multiprocessing.Queue"""
 
     # http://www.jeffknupp.com/blog/2014/02/11/a-celerylike-python-task-queue-in-55-lines-of-code/
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def __init__(
         self,
         work_todo=None,
@@ -506,7 +529,8 @@ class TaskMgr(object):
             self.controller = queue
             self.supervise()  # hot_loops automatically supervise()
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def validate_attribute_types(self):
 
         assert isinstance(self.worker_count, int)
@@ -523,7 +547,8 @@ class TaskMgr(object):
 
         return True
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def configure_logging(self):
         """
         Configure logging parameters output to sys.stdout.
@@ -542,7 +567,8 @@ class TaskMgr(object):
                 self.log_level = 0
             logger.enable(PACKAGE_NAME)
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def supervise(self):
         """
         If not in a hot_loop, call supervise() to start the tasks.
@@ -589,6 +615,19 @@ class TaskMgr(object):
                 task = r_msg.get("task")
                 w_id = r_msg.get("w_id")
                 state = r_msg.get("state", "")
+
+                # Verify that the user called `super().__init__()` in their
+                #      BaseTask() subclass __init__()
+                try:
+                    if (task is not None) and (task.called_super_in_task_subclass is True):
+                        # In this case, the user DID call `super().__init__()`
+                        #    in their BaseTask() subclass __init__()
+                        pass
+                except NameError:
+                    # In this case, the user did NOT call `super().__init__()`
+                    #    in their BaseTask() subclass __init__()
+                    raise NameError("You did NOT call `super().__init__()` in your BaseTask() sub-class.")
+
                 if state == "__ACK__":
                     self.worker_assignments[w_id] = task
                     self.work_todo.remove(task)
@@ -692,7 +731,8 @@ class TaskMgr(object):
                 logger.info(stats.log_stats_message)
             return self.retval
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def calc_wait_time(self, exec_times):
         num_samples = float(len(exec_times))
 
@@ -712,7 +752,8 @@ class TaskMgr(object):
 
         return wait_time
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def queue_tasks_from_controller(self, delay=0.0):
         finished = False
         while not finished:
@@ -730,12 +771,14 @@ class TaskMgr(object):
             # except (Exception) as e:
             #    tb.print_exc()
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def queue_task(self, task):
         task.queue_time = time.time()  # Record the queueing time
         self.todo_q_send({"task": task})
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def are_tasks_finished(self):
         if (len(self.work_todo) == 0) and (len(self.worker_assignments.keys()) == 0):
             return True
@@ -744,12 +787,14 @@ class TaskMgr(object):
             return True
         return False
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def stop_workers(self):
         stop = {"state": "__DIE__"}
         [self.todo_q_send(stop) for x in range(0, self.worker_count)]
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def respawn_dead_workers(self):
         """Respawn workers / tasks upon crash"""
         for w_id, p in self.workers.items():
@@ -788,7 +833,8 @@ class TaskMgr(object):
                 self.workers[w_id].daemon = True
                 self.workers[w_id].start()
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def spawn_workers(self):
         workers = {}
         for w_id in range(0, self.worker_count):
@@ -801,7 +847,8 @@ class TaskMgr(object):
             workers[w_id].start()
         return workers
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def todo_q_send(self, msg_dict):
 
         # Check whether msg_dict can be pickled...
@@ -889,7 +936,8 @@ class TaskMgr(object):
                     )
                 )
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def invalid_dict_pickle_keys(self, msg_dict):
         """Return a list of keys that can't be pickled.  Return [] if
         there are no pickling problems with the values associated with the
@@ -907,7 +955,8 @@ class TaskMgr(object):
                 no_pickle_keys.append(key)  # This key has an unpicklable value
         return no_pickle_keys
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def invalid_obj_pickle_attrs(self, thisobj):
         no_pickle_attrs = []
         for attr, val in vars(thisobj).items():
@@ -925,7 +974,8 @@ class TaskMgr(object):
 class ControllerQueue(object):
     """A set of queues to manage a continuous hot TaskMgr work loop"""
 
-    @logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    #@logger.catch(default=True, onerror=lambda _: sys.exit(1))
+    @logger.catch(reraise=True)
     def __init__(self):
         ## to and from are with respect to the (client) controller object
         self.from_taskmgr_q = py3_mp_queue()  # sent to the controller from TaskMgr
