@@ -6,6 +6,11 @@ class BaseTask(object):
     __metaclass__ = ABCMeta
 
     def __init__(self):
+
+        # Remind user that they forgot to call super().__init__() in
+        # their inherited class __init__().
+        self.called_super_in_task_subclass = True
+
         self.addr = None  # String IP Address
         self.task_stop = 0.0  # Data timestamp, float seconds since epoch
         self.task_start = 0.0  # When task was run by a worker
@@ -13,6 +18,7 @@ class BaseTask(object):
 
         self.worker_loop_delay = 0.00001  # Default 10us worker sleep delay
 
+
     def __hash__(self):
         return id(self)
 
@@ -30,4 +36,5 @@ class BaseTask(object):
     @abstractmethod
     def run(self):
         """Define what should be done"""
+
         raise NotImplementedError
